import warnings

import polars as pl
from sentier_data_tools import GeonamesIRI

BASE_URL = "https://sws.geonames.org/"
LOOKUP_FILE = "../data/iso_iri_table.csv"
LOOKUP_SEPARATOR = "\t"


def get_geonames_iri_from_iso_code(iso_code: str) -> GeonamesIRI:
    """
    Lookup the geoname identifier for a given ISO code using a CSV lookup table.

    Parameters
    ----------
    iso_code : str
        The ISO country code. It can be either a two-digit ISO 3166-1 alpha-2 code
        or a three-digit ISO 3166-1 alpha-3 code.

    Returns
    -------
    GeonamesIRI
        The IRI of the location identifier.

    Raises
    ------
    ValueError
        If the `iso_code` is not two or three characters long or if no matching geoname identifier is found.
    """
    if len(iso_code) not in [2, 3]:
        raise ValueError("Please provide a two-digit or three-digit ISO code")

    lookup = pl.scan_csv(LOOKUP_FILE, separator=LOOKUP_SEPARATOR)
    column = "ISO3" if len(iso_code) == 3 else "ISO2"
    query = f"SELECT geonameid FROM self WHERE {column} = '{iso_code}'"
    res = lookup.sql(query).collect()

    if res.is_empty():
        warnings.warn(f"ISO code {iso_code} not found")
        return None

    geoname_id = res.item(0, 0)
    iri = BASE_URL + str(geoname_id) + "/"
    return GeonamesIRI(iri)



import warnings
from datetime import datetime

import pandas as pd
from loguru import logger
from sentier_data_tools import (
    Dataset,
    Demand,
    Flow,
    GeonamesIRI,
    ProductIRI,
    RunConfig,
    SentierModel,
    UnitIRI,
)

from sentier_peakachu.iri_mapping import (
    DIRTY_BONSAI_PRODUCT_IRIS_MAPPING,
    DIRTY_TRACE_AGGREGATION,
)

from .utils_time import filter_timespan


class TopModel(SentierModel):
    pass


class BottomModel(SentierModel):
    pass


class ElectricitySourceModel(BottomModel):
    provides = {
        ProductIRI("https://example.com/pv"): "pv",
        ProductIRI("https://example.com/coa"): "coal",
        ProductIRI("https://example.com/gas"): "gas",
    }
    aliases = {}

    def run(self, abbreviate_iris: bool = True) -> tuple[list[Demand], list[Flow]]:
        self.prepare()

    def prepare(self) -> None:
        self.get_model_data()
        self.data_validity_checks()
        self.resample()

    def get_model_data(self) -> pd.DataFrame:
        datasets = list(
            Dataset.select().where(
                (Dataset.product == self.demand.product_iri)
                & (Dataset.location == self.demand.spatial_context)
            )
        )
        if not datasets:
            raise ValueError(
                f"No plant emissions data available for {self.demand.product_iri} in {self.demand.spatial_context}."
            )
        df_merged = self.merge_datasets_to_dataframes(datasets)
        return filter_timespan(df_merged, self.demand.begin_date, self.demand.end_date)

    def calculate_average_source_emission_factor(
        self, df: pd.DataFrame
    ) -> pd.DataFrame:
        if df.empty:
            raise ValueError(f"Input DataFrame is empty.")

        if df.empty:
            raise ValueError(
                f"No data available for the specific time range between {self.demand.begin_date} and {self.demand.end_date}."
            )

        total_emissions = df[
            "http://openenergy-platform.org/ontology/oeo/OEO_00260007"
        ].sum()
        total_production = df["https://example.com/process-terms/powergeneration"].sum()
        return total_emissions / total_production


class ElectricityMixModel(TopModel):
    provides = {
        ProductIRI(
            "http://openenergy-platform.org/ontology/oeo/OEO_00000139"
        ): "electricity",
    }
    aliases = {}

    def run(self, abbreviate_iris: bool = True) -> tuple[list[Demand], list[Flow]]:
        self.prepare()

    def prepare(self) -> None:
        self.get_model_data()
        self.data_validity_checks()
        self.resample()

    def get_model_data(self) -> pd.DataFrame:
        datasets = list(
            Dataset.select().where(
                (Dataset.product == self.demand.product_iri)
                & (Dataset.location == self.demand.spatial_context)
            )
        )
        if not datasets:
            raise ValueError(
                f"No electricity mix data available for {self.demand.product_iri} in {self.demand.spatial_context}."
            )
        df_merged = self.merge_datasets_to_dataframes(datasets)
        # logger.debug(f"Electricity mix data before filtering: {df_merged}")
        rounded_begin_date = pd.Timestamp(self.demand.begin_date).round("15min")
        rounded_end_date = pd.Timestamp(self.demand.end_date).round("15min")

        df_merged["https://vocab.sentier.dev/units/quantity-kind/Time"] = (
            pd.to_datetime(
                df_merged["https://vocab.sentier.dev/units/quantity-kind/Time"]
            ).dt.tz_localize(None)
        )
        filtered_df = df_merged[
            (
                df_merged["https://vocab.sentier.dev/units/quantity-kind/Time"]
                >= rounded_begin_date
            )
            & (
                df_merged["https://vocab.sentier.dev/units/quantity-kind/Time"]
                <= rounded_end_date
            )
        ]
        return filtered_df

    def calculate_market_mix(self, df_mixes: pd.DataFrame) -> pd.DataFrame:
        if df_mixes.empty:
            raise ValueError(f"Input DataFrame is empty.")
        average_mix = df_mixes.mean().drop(
            "https://vocab.sentier.dev/units/quantity-kind/Time"
        )
        return average_mix / average_mix.sum()

    def calculate_impact(self, market_shares: pd.Series) -> float:
        emission_factors = []
        market_shares.index = market_shares.index.map(
            lambda x: DIRTY_TRACE_AGGREGATION.get(x, x)
        )  # aggregate specific technologies
        grouped_market_shares = market_shares.groupby(market_shares.index).sum()

        considered_technologies_in_market = []
        for technology in grouped_market_shares.index:
            if (
                technology in DIRTY_TRACE_AGGREGATION.values()
            ):  # fossile technologies form climate traxe
                ef = self.calculate_technology_emission_factor(ProductIRI(technology))
                emission_factors.append(ef)
                considered_technologies_in_market.append(technology)
            elif technology in DIRTY_BONSAI_PRODUCT_IRIS_MAPPING.values():
                ef = self.get_bonsai_emission_factor(technology)
                if ef is None:
                    continue
                emission_factors.append(ef)
                considered_technologies_in_market.append(technology)
            else:
                warnings.warn(f"No emission factor available for {technology}.")

        considered_market_shares = grouped_market_shares.loc[
            considered_technologies_in_market
        ]
        considered_market_shares_normalized = (
            considered_market_shares / considered_market_shares.sum()
        )

        logger.debug(f"Market shares: {considered_market_shares_normalized}")
        logger.debug(f"Emission factors: {emission_factors}")
        return self.demand.amount * sum(
            considered_market_shares_normalized * emission_factors
        )

    def get_bonsai_emission_factor(self, product_iri: ProductIRI) -> float:
        datasets = list(
            Dataset.select().where(
                (Dataset.product == product_iri)
                & (Dataset.location == self.demand.spatial_context)
            )
        )
        if len(datasets) == 0:
            return None
        if len(datasets) == 1:
            dataframe = datasets[0].dataframe
            total_emissions = sum(
                dataframe["https://example.com/direct_CO2_emissions"],
                dataframe["https://example.com/indirect_CO2_emissions"],
            )
        else:
            merged_df = self.merge_datasets_to_dataframes(datasets)
            total_emissions = sum(
                merged_df["https://example.com/direct_CO2_emissions"].mean(),
                merged_df["https://example.com/indirect_CO2_emissions"].mean(),
            )

        logger.debug(f"{product_iri}: {total_emissions} kgCO2eq/kWh")
        return float(total_emissions)

    def calculate_technology_emission_factor(self, product_iri: ProductIRI) -> float:
        demand = Demand(
            product_iri=product_iri,
            unit_iri=UnitIRI("https://vocab.sentier.dev/units/unit/KilowattHour"),
            amount=1,
            spatial_context=self.demand.spatial_context,
            begin_date=self.demand.begin_date,
            end_date=self.demand.end_date,
        )
        model = ElectricitySourceModel(demand=demand, run_config=RunConfig())
        plant_emissions = model.get_model_data()
        return model.calculate_average_source_emission_factor(plant_emissions)



ENTSOE_PRODUCT_IRIS_MAPPING = {
    "Fossil Oil": "https://vocab.sentier.dev/products/electricity/crude-oil",
    "Fossil Oil shale": "https://vocab.sentier.dev/products/electricity/oil-shale",
    "Fossil Gas": "https://vocab.sentier.dev/products/electricity/natural-gas",
    "Fossil Coal-derived gas": "https://vocab.sentier.dev/products/electricity/coalbed-gas",
    "Fossil Hard coal": "https://vocab.sentier.dev/products/electricity/hard-coal",
    "Fossil Brown coal/Lignite": "https://vocab.sentier.dev/products/electricity/lignite",
    "Fossil Peat": "https://vocab.sentier.dev/products/electricity/peat",
    "Wind Offshore": "https://vocab.sentier.dev/products/electricity/offshore",
    "Wind Onshore": "https://vocab.sentier.dev/products/electricity/onshore",
    "Hydro Water Reservoir": "https://vocab.sentier.dev/products/electricity/hydro-reservoirs",
    "Hydro Pumped Storage": "https://vocab.sentier.dev/products/electricity/pumped-storage",
    "Hydro Run-of-river and poundage": "https://vocab.sentier.dev/products/electricity/run-of-river",
    "Marine": "https://vocab.sentier.dev/products/electricity/marine",
    "Solar": "http://openenergyplatform.org/ontology/oeo/OEO_00010419/",
    "Geothermal": "https://vocab.sentier.dev/products/electricity/geothermal",
    "Biomass": "https://vocab.sentier.dev/products/electricity/biomass",
    "Other renewable": "https://vocab.sentier.dev/products/electricity/other-renewable-sources",
    "Nuclear": "http://openenergyplatform.org/ontology/oeo/OEO_00010417/",
    "Waste": "https://vocab.sentier.dev/products/electricity/waste",
    "Other": "https://vocab.sentier.dev/products/electricity/other",
}

DIRTY_TRACE_AGGREGATION = {
    "https://vocab.sentier.dev/products/electricity/crude-oil": "https://vocab.sentier.dev/products/electricity/fossil-oil",
    "https://vocab.sentier.dev/products/electricity/oil-shale": "https://vocab.sentier.dev/products/electricity/fossil-oil",
    "https://vocab.sentier.dev/products/electricity/natural-gas": "https://vocab.sentier.dev/products/electricity/fossil-gas",
    "https://vocab.sentier.dev/products/electricity/coalbed-gas": "https://vocab.sentier.dev/products/electricity/fossil-gas",
    "https://vocab.sentier.dev/products/electricity/hard-coal": "https://vocab.sentier.dev/products/electricity/fossil-coal",
    "https://vocab.sentier.dev/products/electricity/lignite": "https://vocab.sentier.dev/products/electricity/fossil-coal",
    "https://vocab.sentier.dev/products/electricity/peat": "https://vocab.sentier.dev/products/electricity/peat",
    # "https://vocab.sentier.dev/products/electricity/offshore": None,
    # "https://vocab.sentier.dev/products/electricity/onshore": None,
    # "https://vocab.sentier.dev/products/electricity/hydro-reservoirs": None,
    # "https://vocab.sentier.dev/products/electricity/pumped-storage": None,
    # "https://vocab.sentier.dev/products/electricity/run-of-river": None,
    # "https://vocab.sentier.dev/products/electricity/marine": None,
    # "http://openenergyplatform.org/ontology/oeo/OEO_00010419/": None,
    # "https://vocab.sentier.dev/products/electricity/geothermal": None,
    # "https://vocab.sentier.dev/products/electricity/biomass": None,
    # "https://vocab.sentier.dev/products/electricity/other-renewable-sources": None,
    # "http://openenergyplatform.org/ontology/oeo/OEO_00010417/": None,
    # "https://vocab.sentier.dev/products/electricity/waste": "https://vocab.sentier.dev/products/electricity/waste",
    # "https://vocab.sentier.dev/products/electricity/other": None,
}

TRACE_PRODUCT_IRIS_MAPPING = {
    "oil": "https://vocab.sentier.dev/products/electricity/fossil-oil",
    "gas": "https://vocab.sentier.dev/products/electricity/fossil-gas",
    "other_fossil": "https://vocab.sentier.dev/products/electricity/peat",
    "coal": "https://vocab.sentier.dev/products/electricity/fossil-coal",
    "biomass": "https://vocab.sentier.dev/products/electricity/biomass",
    "waste": "https://vocab.sentier.dev/products/electricity/waste",
}

DIRTY_BONSAI_PRODUCT_IRIS_MAPPING = {  # RENEWABLES ONLY
    # "Production of electricity by biomass and waste": "https://vocab.sentier.dev/products/electricity/biomass",
    "Production of electricity by geothermal": "https://vocab.sentier.dev/products/electricity/geothermal",
    "Production of electricity by hydro": "https://vocab.sentier.dev/products/electricity/run-of-river",  # WRONG
    "Production of electricity by nuclear": "http://openenergyplatform.org/ontology/oeo/OEO_00010417/",
    "Production of electricity by solar photovoltaic": "http://openenergyplatform.org/ontology/oeo/OEO_00010419/",
    "Production of electricity by tide, wave, ocean": "https://vocab.sentier.dev/products/electricity/marine",
    "Production of electricity by wind": "https://vocab.sentier.dev/products/electricity/offshore",  # WRONG
}

# DIRTY_BONSAI_AGGREGATION = {



import pandas as pd

from sentier_peakachu.entsoe import get_generation_data


def test_entsoe():
    query = {
        "country_code": "DE",
        "start": pd.Timestamp("20241008", tz="Europe/Brussels"),
        "end": pd.Timestamp("20241009", tz="Europe/Brussels"),
    }

    get_generation_data(**query)



"""sentier_peakachu."""

__all__ = (
    "__version__",
    "ElectricitySourceModel",
    "ElectricityMixModel",
    "create_local_datastorage",
    "get_geonames_iri_from_iso_code",
    "filter_timespan",
)

__version__ = "0.2.0"

from sentier_peakachu.data import create_local_electricity_datastorage
from sentier_peakachu.model import ElectricityMixModel, ElectricitySourceModel
from sentier_peakachu.utils_location import get_geonames_iri_from_iso_code
from sentier_peakachu.utils_time import filter_timespan



import warnings
from datetime import datetime

import pandas as pd
import sentier_data_tools as sdt

from sentier_peakachu.entsoe import get_generation_data
from sentier_peakachu.iri_mapping import (
    DIRTY_BONSAI_PRODUCT_IRIS_MAPPING,
    DIRTY_TRACE_AGGREGATION,
    ENTSOE_PRODUCT_IRIS_MAPPING,
    TRACE_PRODUCT_IRIS_MAPPING,
)
from sentier_peakachu.utils_location import get_geonames_iri_from_iso_code


def create_local_electricity_datastorage(reset: bool = True):
    if reset:
        sdt.reset_local_database()
    start_time = pd.Timestamp("20221008", tz="Europe/Brussels")
    end_time = pd.Timestamp("20221009", tz="Europe/Brussels")
    create_country_mix_dataset("DE", start_time, end_time)

    create_plant_emission_datasets()
    create_bonsai_emission_factor_datasets()


def create_country_mix_dataset(
    country_code: str, start_time: pd.Timestamp, end_time: pd.Timestamp
):
    # DF1

    metadata = sdt.Datapackage(
        name="electricity_markets",
        description="Electricity markets data from ENTSO-E",
        contributors=[
            {
                "title": "Peakachu",
                "path": "https://github.com/TimoDiepers/sentier_peakachu/",
                "role": "author",
            },
        ],
        homepage="https://github.com/TimoDiepers/sentier_peakachu/",
    ).metadata()

    df = get_generation_data(
        country_code=country_code,
        start=start_time,
        end=end_time,
    )
    df.index.name = "https://vocab.sentier.dev/units/quantity-kind/Time"
    df = df.reset_index()

    df = df.rename(columns=ENTSOE_PRODUCT_IRIS_MAPPING)

    units_tech = ["https://vocab.sentier.dev/units/unit/MegaW-HR"] * len(
        ENTSOE_PRODUCT_IRIS_MAPPING
    )  # MW not Mwh but no entry in SKOSMOS
    units_time = [
        "https://vocab.sentier.dev/units/unit/SEC",
    ]

    UNITS = units_time + units_tech

    sdt.Dataset(
        name="electricity mixes",
        dataframe=df,
        kind=sdt.DatasetKind.BOM,
        product="http://openenergy-platform.org/ontology/oeo/OEO_00000139",
        columns=[{"iri": x, "unit": y} for x, y in zip(df.columns, UNITS)],
        metadata=metadata,
        location=get_geonames_iri_from_iso_code(country_code),
        version=1,
        valid_from=datetime(2018, 1, 1),
        valid_to=datetime(2028, 1, 1),
    ).save()


def create_plant_emission_datasets():
    # DF2
    metadata = sdt.Datapackage(
        name="emission data power plants",
        description="Climate trace emission data for power plants",
        contributors=[
            {
                "title": "Peakachu",
                "path": "https://github.com/TimoDiepers/sentier_peakachu/",
                "role": "author",
            },
        ],
        homepage="https://github.com/TimoDiepers/sentier_peakachu/",
    ).metadata()

    COLUMNS_POWERPLANTS = [
        "https://example.com/model-terms/identifier",  # to be added to SKOSMOS model terms
        "https://example.com/model-terms/name",  # to be added to SKOSMOS model terms
        "https://example.com/model-terms/start_time",  # to be added to SKOSMOS model terms
        "https://example.com/model-terms/end_time",  # to be added to SKOSMOS model terms
        "https://example.com/process-terms/powergeneration",  # to be added to SKOSMOS process terms
        "http://openenergy-platform.org/ontology/oeo/OEO_00260007",  # CO2 emission
    ]

    UNITS_POWERPLANTS = [
        "https://example.com/model-terms/integer",  # to be added to SKOSMOS model terms
        "https://example.com/model-terms/name",  # to be added to SKOSMOS model terms
        "https://vocab.sentier.dev/units/unit/SEC",
        "https://vocab.sentier.dev/units/unit/SEC",
        "https://vocab.sentier.dev/units/unit/MegaW-HR",
        "https://vocab.sentier.dev/units/unit/TON_Metric",
    ]

    trace_frame = pd.read_csv("../data/electricity-generation_emissions_sources.csv")

    filtered_df = trace_frame[trace_frame["gas"] == "co2e_100yr"]
    grouped_dfs = {
        name: group
        for name, group in filtered_df.groupby(["iso3_country", "source_type"])
    }

    for (country, source_type), df in grouped_dfs.items():
        if source_type not in TRACE_PRODUCT_IRIS_MAPPING.keys():
            warnings.warn(
                f"Source type {source_type} not found, skipping Dataset creation"
            )
            continue

        geonames_iri = get_geonames_iri_from_iso_code(country)
        if not geonames_iri:
            warnings.warn(
                f"Location not found for {country}, skipping Dataset creation"
            )
            continue
        filtered_df = df[
            [
                "source_id",
                "source_name",
                "start_time",
                "end_time",
                "activity",
                "emissions_quantity",
            ]
        ]
        filtered_df.columns = COLUMNS_POWERPLANTS
        valid_from_str = min(df["start_time"])
        valid_to_str = max(df["end_time"])

        sdt.Dataset(
            name=f"power plant data, {country}, {source_type}",
            dataframe=filtered_df,
            kind=sdt.DatasetKind.BOM,
            product=TRACE_PRODUCT_IRIS_MAPPING[source_type],
            columns=[
                {"iri": x, "unit": y} for x, y in zip(df.columns, UNITS_POWERPLANTS)
            ],
            metadata=metadata,
            location=geonames_iri,
            version=1,
            valid_from=datetime.strptime(valid_from_str, "%Y-%m-%d %H:%M:%S"),
            valid_to=datetime.strptime(valid_to_str, "%Y-%m-%d %H:%M:%S"),
        ).save()


def create_bonsai_emission_factor_datasets():
    """
    Create datasets for emission factors for different sources of electricity from the bonsai database,
    splitting emissions in direct and indirect emissions.
    Unit is kg CO2-eq/kWh.
    """

    metadata = sdt.Datapackage(
        name="emission factors for regional electricity producing technologies",
        description="Bonsai emission factor data for regional electricity producing technologies",
        contributors=[
            {
                "title": "Karin Treyer",
                "path": "https://www.psi.ch/en/ta/people/karin-treyer",
                "role": "author",
            },
            {
                "title": "Chris Mutel",
                "path": "https://chris.mutel.org/",
                "role": "wrangler",
            },
        ],
        homepage="https://example.com/additional_inventories",
    ).metadata()

    UNITS_EMISSION_FACTORS = [
        "https://example.com/units/kgCO2eqPerkWh",
        "https://example.com/units/kgCO2eqPerkWh",
    ]
    COLUMNS_EMISSION_FACTORS = [
        "https://example.com/direct_CO2_emissions",
        "https://example.com/indirect_CO2_emissions",
    ]

    bonsai_frame = pd.read_csv("../data/bonsai_emission_factors.csv", delimiter=";")

    filtered_df = bonsai_frame[
        [
            "description",
            "region_code",
            "direct emission factor",
            "indirect emission factor",
        ]
    ]

    grouped_dfs = {
        name: group
        for name, group in filtered_df.groupby(["region_code", "description"])
    }

    for (country, technology), df in grouped_dfs.items():
        geonames_iri = get_geonames_iri_from_iso_code(country)
        if not geonames_iri:
            warnings.warn(
                f"Location not found for {country}, skipping Dataset creation"
            )
            continue
        technology_iri = DIRTY_BONSAI_PRODUCT_IRIS_MAPPING.get(technology)
        if not technology_iri:
            warnings.warn(
                f"Technology {technology} not found, skipping Dataset creation"
            )
            continue
        df = df[["direct emission factor", "indirect emission factor"]]
        df.columns = COLUMNS_EMISSION_FACTORS
        valid_from_str = "2016-01-01"  # Bonsai/EXIOBASE data from year 2016
        valid_to_str = "2024-12-31"

        sdt.Dataset(
            name=f"bonsai emission factors, {country}, {technology}",
            dataframe=df,
            kind=sdt.DatasetKind.BOM,
            product=technology_iri,
            columns=[
                {"iri": x, "unit": y}
                for x, y in zip(df.columns, UNITS_EMISSION_FACTORS)
            ],
            metadata=metadata,
            location=geonames_iri,
            version=1,
            valid_from=valid_from_str,
            valid_to=valid_to_str,
        ).save()



from datetime import datetime

import pandas as pd


def filter_timespan(
    df: pd.DataFrame, begin_date: datetime, end_date: datetime
) -> pd.DataFrame:
    return df[
        (
            df["https://example.com/model-terms/start_time"]
            <= begin_date.strftime("%Y-%m-%d %H:%M:%S")
        )
        & (
            df["https://example.com/model-terms/end_time"]
            >= end_date.strftime("%Y-%m-%d %H:%M:%S")
        )
    ]



import os

import pandas as pd
from dotenv import load_dotenv
from entsoe import EntsoePandasClient


def get_api_key():
    """
    Getting the API key from the .env file.
    """
    load_dotenv()
    API_KEY = os.getenv("ENTSOE_API_KEY")
    return API_KEY


def filter_actual_aggregated(df):
    """
    Curates the generation data from ENTSO-E API to only include the actual aggregated generation
    data.
    """
    if df.columns.nlevels == 1:  # sometimes, consumption data is not available
        return df

    df_filtered = df.loc[:, df.columns.get_level_values(1) == "Actual Aggregated"]
    df_filtered.columns = df_filtered.columns.droplevel(1)
    return df_filtered


def get_generation_data(country_code: str, start: pd.Timestamp, end: pd.Timestamp):
    """
    Get generation data from ENTSO-E API for a specific country and time range.

    Parameters
    ----------
    country_code : str
        Country code (e.g. 'DE' for Germany).
    start : pd.Timestamp
        Start date and time.
    end : pd.Timestamp
        End date and time.

    Returns
    -------
    pd.DataFrame
        Generation data for the specified country and time range.
    """
    client = EntsoePandasClient(api_key=get_api_key())
    df = client.query_generation(country_code=country_code, start=start, end=end)
    return filter_actual_aggregated(df)
